#    Copyright (c) 2014 - 2019 StorPool
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""StorPool block device driver"""

import platform

from os_brick.initiator import storpool_utils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import units

from cinder.common import constants
from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume import volume_types

LOG = logging.getLogger(__name__)


storpool_opts = [
    cfg.StrOpt('storpool_template',
               default=None,
               help='The StorPool template for volumes with no type.'),
    cfg.IntOpt('storpool_replication',
               default=3,
               help='The default StorPool chain replication value.  '
                    'Used when creating a volume with no specified type if '
                    'storpool_template is not set.  Also used for calculating '
                    'the apparent free space reported in the stats.'),
]

CONF = cfg.CONF
CONF.register_opts(storpool_opts, group=configuration.SHARED_CONF_GROUP)


class StorPoolConfigurationInvalid(exception.CinderException):
    message = _("Invalid parameter %(param)s in the %(section)s section "
                "of the /etc/storpool.conf file: %(error)s")


@interface.volumedriver
class StorPoolDriver(driver.VolumeDriver):
    """The StorPool block device driver.

    Version history:

    .. code-block:: none

        0.1.0   - Initial driver
        0.2.0   - Bring the driver up to date with Kilo and Liberty:
                  - implement volume retyping and migrations
                  - use the driver.*VD ABC metaclasses
                  - bugfix: fall back to the configured StorPool template
        1.0.0   - Imported into OpenStack Liberty with minor fixes
        1.1.0   - Bring the driver up to date with Liberty and Mitaka:
                  - drop the CloneableVD and RetypeVD base classes
                  - enable faster volume copying by specifying
                    sparse_volume_copy=true in the stats report
        1.1.1   - Fix the internal _storpool_client_id() method to
                  not break on an unknown host name or UUID; thus,
                  remove the StorPoolConfigurationMissing exception.
        1.1.2   - Bring the driver up to date with Pike: do not
                  translate the error messages
        1.2.0   - Inherit from VolumeDriver, implement get_pool()
        1.2.1   - Implement interface.volumedriver, add CI_WIKI_NAME,
                  fix the docstring formatting
        1.2.2   - Reintroduce the driver into OpenStack Queens,
                  add ignore_errors to the internal _detach_volume() method
        1.2.3   - Advertise some more driver capabilities.
        2.0.0   - Implement revert_to_snapshot().
        2.1.0   - Use the new API client in os-brick to communicate with the
                  StorPool API instead of packages `storpool` and
                  `storpool.spopenstack`

    """

    VERSION = '2.1.0'
    CI_WIKI_NAME = 'StorPool_distributed_storage_CI'

    def __init__(self, *args, **kwargs):
        super(StorPoolDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(storpool_opts)
        self._sp_config = None
        self._ourId = None
        self._ourIdInt = None
        self._sp_api = None
        self._volume_prefix = None

    @staticmethod
    def get_driver_options():
        return storpool_opts

    def _backendException(self, e):
        return exception.VolumeBackendAPIException(data=str(e))

    def _template_from_volume(self, volume):
        default = self.configuration.storpool_template
        vtype = volume['volume_type']
        if vtype is not None:
            specs = volume_types.get_volume_type_extra_specs(vtype['id'])
            if specs is not None:
                return specs.get('storpool_template', default)
        return default

    def get_pool(self, volume):
        template = self._template_from_volume(volume)
        if template is None:
            return 'default'
        else:
            return 'template_' + template

    def create_volume(self, volume):
        size = int(volume['size']) * units.Gi
        name = storpool_utils.os_to_sp_volume_name(
            self._volume_prefix, volume['id'])
        template = self._template_from_volume(volume)

        create_request = {'name': name, 'size': size}
        if template is not None:
            create_request['template'] = template
        else:
            create_request['replication'] = \
                self.configuration.storpool_replication

        try:
            self._sp_api.volume_create(create_request)
        except storpool_utils.StorPoolAPIError as e:
            raise self._backendException(e)

    def _storpool_client_id(self, connector):
        hostname = connector['host']
        if hostname == self.host or hostname == CONF.host:
            hostname = platform.node()
        try:
            cfg = storpool_utils.get_conf(section=hostname)
            return int(cfg['SP_OURID'])
        except KeyError:
            return 65
        except Exception as e:
            raise StorPoolConfigurationInvalid(
                section=hostname, param='SP_OURID', error=e)

    def validate_connector(self, connector):
        return self._storpool_client_id(connector) >= 0

    def initialize_connection(self, volume, connector):
        return {'driver_volume_type': 'storpool',
                'data': {
                    'client_id': self._storpool_client_id(connector),
                    'volume': volume['id'],
                    'access_mode': 'rw',
                }}

    def terminate_connection(self, volume, connector, **kwargs):
        pass

    def create_snapshot(self, snapshot):
        volname = storpool_utils.os_to_sp_volume_name(
            self._volume_prefix, snapshot['volume_id'])
        name = storpool_utils.os_to_sp_snapshot_name(
            self._volume_prefix, 'snap', snapshot['id'])
        try:
            self._sp_api.snapshot_create(volname, {'name': name})
        except storpool_utils.StorPoolAPIError as e:
            raise self._backendException(e)

    def create_volume_from_snapshot(self, volume, snapshot):
        size = int(volume['size']) * units.Gi
        volname = storpool_utils.os_to_sp_volume_name(
            self._volume_prefix, volume['id'])
        name = storpool_utils.os_to_sp_snapshot_name(
            self._volume_prefix, 'snap', snapshot['id'])
        try:
            self._sp_api.volume_create({
                'name': volname,
                'size': size,
                'parent': name
            })
        except storpool_utils.StorPoolAPIError as e:
            raise self._backendException(e)

    def create_cloned_volume(self, volume, src_vref):
        refname = storpool_utils.os_to_sp_volume_name(
            self._volume_prefix, src_vref['id'])
        size = int(volume['size']) * units.Gi
        volname = storpool_utils.os_to_sp_volume_name(
            self._volume_prefix, volume['id'])

        src_volume = self.db.volume_get(
            context.get_admin_context(),
            src_vref['id'],
        )
        src_template = self._template_from_volume(src_volume)

        template = self._template_from_volume(volume)
        LOG.debug('clone volume id %(vol_id)r template %(template)r', {
            'vol_id': volume['id'],
            'template': template,
        })
        if template == src_template:
            LOG.info('Using baseOn to clone a volume into the same template')
            try:
                self._sp_api.volume_create({
                    'name': volname,
                    'size': size,
                    'baseOn': refname,
                })
            except storpool_utils.StorPoolAPIError as e:
                raise self._backendException(e)

            return None

        snapname = storpool_utils.os_to_sp_snapshot_name(
            self._volume_prefix, 'clone', volume['id'])
        LOG.info(
            'A transient snapshot for a %(src)s -> %(dst)s template change',
            {'src': src_template, 'dst': template})
        try:
            self._sp_api.snapshot_create(refname, {'name': snapname})
        except storpool_utils.StorPoolAPIError as e:
            if e.name != 'objectExists':
                raise self._backendException(e)

        try:
            try:
                self._sp_api.snapshot_update(
                    snapname,
                    {'template': template},
                )
            except storpool_utils.StorPoolAPIError as e:
                raise self._backendException(e)

            try:
                self._sp_api.volume_create({
                    'name': volname,
                    'size': size,
                    'parent': snapname
                })
            except storpool_utils.StorPoolAPIError as e:
                raise self._backendException(e)

            try:
                self._sp_api.snapshot_update(
                    snapname,
                    {'tags': {'transient': '1.0'}},
                )
            except storpool_utils.StorPoolAPIError as e:
                raise self._backendException(e)
        except Exception:
            with excutils.save_and_reraise_exception():
                try:
                    LOG.warning(
                        'Something went wrong, removing the transient snapshot'
                    )
                    self._sp_api.snapshot_delete(snapname)
                except storpool_utils.StorPoolAPIError as e:
                    LOG.error(
                        'Could not delete the %(name)s snapshot: %(err)s',
                        {'name': snapname, 'err': str(e)}
                    )

    def create_export(self, context, volume, connector):
        pass

    def remove_export(self, context, volume):
        pass

    def delete_volume(self, volume):
        name = storpool_utils.os_to_sp_volume_name(
            self._volume_prefix, volume['id'])
        try:
            self._sp_api.volumes_reassign([{"volume": name, "detach": "all"}])
            self._sp_api.volume_delete(name)
        except storpool_utils.StorPoolAPIError as e:
            if e.name == 'objectDoesNotExist':
                pass
            else:
                raise self._backendException(e)

    def delete_snapshot(self, snapshot):
        name = storpool_utils.os_to_sp_snapshot_name(
            self._volume_prefix, 'snap', snapshot['id'])
        try:
            self._sp_api.volumes_reassign(
                [{"snapshot": name, "detach": "all"}])
            self._sp_api.snapshot_delete(name)
        except storpool_utils.StorPoolAPIError as e:
            if e.name == 'objectDoesNotExist':
                pass
            else:
                raise self._backendException(e)

    def check_for_setup_error(self):
        try:
            self._sp_config = storpool_utils.get_conf()
            self._sp_api = storpool_utils.StorPoolAPI(
                self._sp_config["SP_API_HTTP_HOST"],
                self._sp_config["SP_API_HTTP_PORT"],
                self._sp_config["SP_AUTH_TOKEN"])
            self._volume_prefix = self._sp_config.get(
                "SP_OPENSTACK_VOLUME_PREFIX", "os")
        except Exception as e:
            LOG.error("StorPoolDriver API initialization failed: %s", e)
            raise

    def _update_volume_stats(self):
        try:
            dl = self._sp_api.disks_list()
            templates = self._sp_api.volume_templates_list()
        except storpool_utils.StorPoolAPIError as e:
            raise self._backendException(e)
        total = 0
        used = 0
        free = 0
        agSize = 512 * units.Mi
        for (id, desc) in dl.items():
            if desc['generationLeft'] != -1:
                continue
            total += desc['agCount'] * agSize
            used += desc['agAllocated'] * agSize
            free += desc['agFree'] * agSize * 4096 / (4096 + 128)

        # Report the free space as if all new volumes will be created
        # with StorPool replication 3; anything else is rare.
        free /= self.configuration.storpool_replication

        space = {
            'total_capacity_gb': total / units.Gi,
            'free_capacity_gb': free / units.Gi,
            'reserved_percentage': 0,
            'multiattach': True,
            'QoS_support': False,
            'thick_provisioning_support': False,
            'thin_provisioning_support': True,
        }

        pools = [dict(space, pool_name='default')]

        pools += [dict(space,
                       pool_name='template_' + t['name'],
                       storpool_template=t['name']
                       ) for t in templates]

        self._stats = {
            # Basic driver properties
            'volume_backend_name': self.configuration.safe_get(
                'volume_backend_name') or 'storpool',
            'vendor_name': 'StorPool',
            'driver_version': self.VERSION,
            'storage_protocol': constants.STORPOOL,
            # Driver capabilities
            'clone_across_pools': True,
            'sparse_copy_volume': True,
            # The actual pools data
            'pools': pools
        }

    def extend_volume(self, volume, new_size):
        size = int(new_size) * units.Gi
        name = storpool_utils.os_to_sp_volume_name(
            self._volume_prefix, volume['id'])
        try:
            self._sp_api.volume_update(name, {'size': size})
        except storpool_utils.StorPoolAPIError as e:
            raise self._backendException(e)

    def ensure_export(self, context, volume):
        # Already handled by Nova's AttachDB, we hope.
        # Maybe it should move here, but oh well.
        pass

    def retype(self, context, volume, new_type, diff, host):
        update = {}

        if diff['encryption']:
            LOG.error('Retype of encryption type not supported.')
            return False

        templ = self.configuration.storpool_template
        repl = self.configuration.storpool_replication
        if diff['extra_specs']:
            # Check for the StorPool extra specs. We intentionally ignore any
            # other extra_specs because the cinder scheduler should not even
            # call us if there's a serious mismatch between the volume types.
            if diff['extra_specs'].get('volume_backend_name'):
                v = diff['extra_specs'].get('volume_backend_name')
                if v[0] != v[1]:
                    # Retype of a volume backend not supported yet,
                    # the volume needs to be migrated.
                    return False
            if diff['extra_specs'].get('storpool_template'):
                v = diff['extra_specs'].get('storpool_template')
                if v[0] != v[1]:
                    if v[1] is not None:
                        update['template'] = v[1]
                    elif templ is not None:
                        update['template'] = templ
                    else:
                        update['replication'] = repl

        if update:
            name = storpool_utils.os_to_sp_volume_name(
                self._volume_prefix, volume['id'])
            try:
                self._sp_api.volume_update(name, **update)
            except storpool_utils.StorPoolAPIError as e:
                raise self._backendException(e)

        return True

    def update_migrated_volume(self, context, volume, new_volume,
                               original_volume_status):
        orig_id = volume['id']
        orig_name = storpool_utils.os_to_sp_volume_name(
            self._volume_prefix, orig_id)
        temp_id = new_volume['id']
        temp_name = storpool_utils.os_to_sp_volume_name(
            self._volume_prefix, temp_id)
        vols = {v['name']: True for v in self._sp_api.volumes_list()}
        if temp_name not in vols:
            LOG.error('StorPool update_migrated_volume(): it seems '
                      'that the StorPool volume "%(tid)s" was not '
                      'created as part of the migration from '
                      '"%(oid)s".', {'tid': temp_id, 'oid': orig_id})
            return {'_name_id': new_volume['_name_id'] or new_volume['id']}

        if orig_name in vols:
            LOG.debug('StorPool update_migrated_volume(): both '
                      'the original volume "%(oid)s" and the migrated '
                      'StorPool volume "%(tid)s" seem to exist on '
                      'the StorPool cluster.',
                      {'oid': orig_id, 'tid': temp_id})
            int_name = temp_name + '--temp--mig'
            LOG.debug('Trying to swap volume names, intermediate "%(int)s"',
                      {'int': int_name})
            try:
                LOG.debug('- rename "%(orig)s" to "%(int)s"',
                          {'orig': orig_name, 'int': int_name})
                self._sp_api.volume_update(orig_name, {'rename': int_name})

                LOG.debug('- rename "%(temp)s" to "%(orig)s"',
                          {'temp': temp_name, 'orig': orig_name})
                self._sp_api.volume_update(temp_name, {'rename': orig_name})

                LOG.debug('- rename "%(int)s" to "%(temp)s"',
                          {'int': int_name, 'temp': temp_name})
                self._sp_api.volume_update(int_name, {'rename': temp_name})
                return {'_name_id': None}
            except storpool_utils.StorPoolAPIError as e:
                LOG.error('StorPool update_migrated_volume(): '
                          'could not rename a volume: '
                          '%(err)s',
                          {'err': e})
                return {'_name_id': new_volume['_name_id'] or new_volume['id']}

        try:
            self._sp_api.volume_update(temp_name, {'rename': orig_name})
            return {'_name_id': None}
        except storpool_utils.StorPoolAPIError as e:
            LOG.error('StorPool update_migrated_volume(): '
                      'could not rename %(tname)s to %(oname)s: '
                      '%(err)s',
                      {'tname': temp_name, 'oname': orig_name, 'err': e})
            return {'_name_id': new_volume['_name_id'] or new_volume['id']}

    def revert_to_snapshot(self, context, volume, snapshot):
        volname = storpool_utils.os_to_sp_volume_name(
            self._volume_prefix, volume['id'])
        snapname = storpool_utils.os_to_sp_snapshot_name(
            self._volume_prefix, 'snap', snapshot['id'])
        try:
            self._sp_api.volume_revert(volname, {'toSnapshot': snapname})
        except storpool_utils.StorPoolAPIError as e:
            LOG.error('StorPool revert_to_snapshot(): could not revert '
                      'the %(vol_id)s volume to the %(snap_id)s snapshot: '
                      '%(err)s',
                      {'vol_id': volume['id'],
                       'snap_id': snapshot['id'],
                       'err': e})
            raise self._backendException(e)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_config import cfg

from cinder import db

CONF = cfg.CONF


class Target(object, metaclass=abc.ABCMeta):
    """Target object for block storage devices.

    Base class for target object, where target
    is data transport mechanism (target) specific calls.
    This includes things like create targets, attach, detach
    etc.

    Base class here does nothing more than set an executor and db as
    well as force implementation of required methods.

    """
    storage_protocol = None
    SHARED_TARGET_SUPPORT = False
    SECONDARY_IP_SUPPORT = True

    def __init__(self, *args, **kwargs):
        # TODO(stephenfin): Drop this in favour of using 'db' directly
        self.db = db
        self.configuration = kwargs.get('configuration')
        self._root_helper = kwargs.get('root_helper',
                                       'sudo cinder-rootwrap %s' %
                                       CONF.rootwrap_config)

    @abc.abstractmethod
    def ensure_export(self, context, volume, volume_path):
        """Synchronously recreates an export for a volume."""
        pass

    @abc.abstractmethod
    def create_export(self, context, volume, volume_path):
        """Exports a Target/Volume.

        Can optionally return a Dict of changes to
        the volume object to be persisted.
        """
        pass

    @abc.abstractmethod
    def remove_export(self, context, volume):
        """Removes an export for a Target/Volume."""
        pass

    @abc.abstractmethod
    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info."""
        pass

    @abc.abstractmethod
    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector."""
        pass

    @staticmethod
    def are_same_connector(A, B):
        """Whether 2 connectors belong to the same host or not.

        This is used for multi attach volumes, to be able to know when there
        are no more attachments on a given host.

        This is the generic implementation, but specific targets may overwrite
        it. For example iSCSI would check the the "initiator" key instead, and
        NVMe-oF would check the "nqn" key.
        """
        a_host = A.get('host')
        return a_host and (a_host == B.get('host'))

    def extend_target(self, volume):
        """Reinitializes a target after the volume has been extended.

        Most drivers don't need to do anything, but in other cases this may
        cause IO disruption.
        """
        pass



# Copyright (C) 2022, 2024, Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Unit tests for Hitachi HBSD Driver."""

import json
from unittest import mock

from oslo_config import cfg
import requests

from cinder import context as cinder_context
from cinder.db.sqlalchemy import api as sqlalchemy_api
from cinder import exception
from cinder.objects import group_snapshot as obj_group_snap
from cinder.objects import snapshot as obj_snap
from cinder.tests.unit import fake_group
from cinder.tests.unit import fake_group_snapshot
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume import driver
from cinder.volume.drivers.hitachi import hbsd_common
from cinder.volume.drivers.hitachi import hbsd_fc
from cinder.volume.drivers.hitachi import hbsd_rest
from cinder.volume.drivers.hitachi import hbsd_rest_api
from cinder.volume.drivers.hitachi import hbsd_utils
from cinder.volume import volume_types
from cinder.volume import volume_utils
from cinder.zonemanager import utils as fczm_utils

# Configuration parameter values
CONFIG_MAP = {
    'serial': '886000123456',
    'my_ip': '127.0.0.1',
    'rest_server_ip_addr': '172.16.18.108',
    'rest_server_ip_port': '23451',
    'port_id': 'CL1-A',
    'host_grp_name': 'HBSD-0123456789abcdef',
    'host_mode': 'LINUX/IRIX',
    'host_wwn': '0123456789abcdef',
    'target_wwn': '1111111123456789',
    'user_id': 'user',
    'user_pass': 'password',
    'pool_name': 'test_pool',
    'auth_user': 'auth_user',
    'auth_password': 'auth_password',
}

REMOTE_CONFIG_MAP = {
    'serial': '886000456789',
    'my_ip': '127.0.0.1',
    'rest_server_ip_addr': '172.16.18.107',
    'rest_server_ip_port': '334',
    'port_id': 'CL2-B',
    'host_grp_name': 'HBSD-0123456789abcdef',
    'host_mode': 'LINUX/IRIX',
    'host_wwn': '0123456789abcdef',
    'target_wwn': '2222222234567891',
    'user_id': 'remote-user',
    'user_pass': 'remote-password',
    'pool_name': 'remote_pool',
    'auth_user': 'remote_user',
    'auth_password': 'remote_password',
}

# Dummy response for FC zoning device mapping
DEVICE_MAP = {
    'fabric_name': {
        'initiator_port_wwn_list': [CONFIG_MAP['host_wwn']],
        'target_port_wwn_list': [CONFIG_MAP['target_wwn']]}}

REMOTE_DEVICE_MAP = {
    'fabric_name': {
        'initiator_port_wwn_list': [REMOTE_CONFIG_MAP['host_wwn']],
        'target_port_wwn_list': [REMOTE_CONFIG_MAP['target_wwn']]}}

DEFAULT_CONNECTOR = {
    'host': 'host',
    'ip': CONFIG_MAP['my_ip'],
    'wwpns': [CONFIG_MAP['host_wwn']],
    'multipath': False,
}

REMOTE_DEFAULT_CONNECTOR = {
    'host': 'host',
    'ip': REMOTE_CONFIG_MAP['my_ip'],
    'wwpns': [REMOTE_CONFIG_MAP['host_wwn']],
    'multipath': False,
}

CTXT = cinder_context.get_admin_context()

TEST_VOLUME = []
for i in range(8):
    volume = {}
    volume['id'] = '00000000-0000-0000-0000-{0:012d}'.format(i)
    volume['name'] = 'test-volume{0:d}'.format(i)
    volume['volume_type_id'] = '00000000-0000-0000-0000-{0:012d}'.format(i)
    if i == 3 or i == 7:
        volume['provider_location'] = None
    elif i == 4:
        volume['provider_location'] = json.dumps(
            {'pldev': 4, 'sldev': 4,
             'remote-copy': hbsd_utils.MIRROR_ATTR})
    elif i == 5:
        volume['provider_location'] = json.dumps(
            {'pldev': 5, 'sldev': 5,
             'remote-copy': hbsd_utils.MIRROR_ATTR})
    elif i == 6:
        volume['provider_location'] = json.dumps(
            {'pldev': 6, 'sldev': 6,
             'remote-copy': hbsd_utils.MIRROR_ATTR})
    else:
        volume['provider_location'] = '{0:d}'.format(i)
    volume['size'] = 128
    if i == 2 or i == 6:
        volume['status'] = 'in-use'
    elif i == 7:
        volume['status'] = None
    else:
        volume['status'] = 'available'
    volume = fake_volume.fake_volume_obj(CTXT, **volume)
    volume.volume_type = fake_volume.fake_volume_type_obj(CTXT)
    TEST_VOLUME.append(volume)


def _volume_get(context, volume_id):
    """Return predefined volume info."""
    return TEST_VOLUME[int(volume_id.replace("-", ""))]


TEST_SNAPSHOT = []
for i in range(2):
    snapshot = {}
    snapshot['id'] = '10000000-0000-0000-0000-{0:012d}'.format(i)
    snapshot['name'] = 'TEST_SNAPSHOT{0:d}'.format(i)
    snapshot['provider_location'] = '{0:d}'.format(i + 1)
    snapshot['status'] = 'available'
    snapshot['volume_id'] = '00000000-0000-0000-0000-{0:012d}'.format(i)
    snapshot['volume'] = _volume_get(None, snapshot['volume_id'])
    snapshot['volume_name'] = 'test-volume{0:d}'.format(i)
    snapshot['volume_size'] = 128
    snapshot = obj_snap.Snapshot._from_db_object(
        CTXT, obj_snap.Snapshot(),
        fake_snapshot.fake_db_snapshot(**snapshot))
    TEST_SNAPSHOT.append(snapshot)

TEST_GROUP = []
for i in range(2):
    group = {}
    group['id'] = '20000000-0000-0000-0000-{0:012d}'.format(i)
    group['status'] = 'available'
    group = fake_group.fake_group_obj(CTXT, **group)
    TEST_GROUP.append(group)

TEST_GROUP_SNAP = []
group_snapshot = {}
group_snapshot['id'] = '30000000-0000-0000-0000-{0:012d}'.format(0)
group_snapshot['status'] = 'available'
group_snapshot = obj_group_snap.GroupSnapshot._from_db_object(
    CTXT, obj_group_snap.GroupSnapshot(),
    fake_group_snapshot.fake_db_group_snapshot(**group_snapshot))
TEST_GROUP_SNAP.append(group_snapshot)

# Dummy response for REST API
POST_SESSIONS_RESULT = {
    "token": "b74777a3-f9f0-4ea8-bd8f-09847fac48d3",
    "sessionId": 0,
}

REMOTE_POST_SESSIONS_RESULT = {
    "token": "b74777a3-f9f0-4ea8-bd8f-09847fac48d4",
    "sessionId": 0,
}

GET_PORTS_RESULT = {
    "data": [
        {
            "portId": CONFIG_MAP['port_id'],
            "portType": "FIBRE",
            "portAttributes": [
                "TAR",
                "MCU",
                "RCU",
                "ELUN"
            ],
            "fabricMode": True,
            "portConnection": "PtoP",
            "lunSecuritySetting": True,
            "wwn": CONFIG_MAP['target_wwn'],
        },
    ],
}

REMOTE_GET_PORTS_RESULT = {
    "data": [
        {
            "portId": REMOTE_CONFIG_MAP['port_id'],
            "portType": "FIBRE",
            "portAttributes": [
                "TAR",
                "MCU",
                "RCU",
                "ELUN"
            ],
            "fabricMode": True,
            "portConnection": "PtoP",
            "lunSecuritySetting": True,
            "wwn": REMOTE_CONFIG_MAP['target_wwn'],
        },
    ],
}

GET_HOST_WWNS_RESULT = {
    "data": [
        {
            "hostGroupNumber": 0,
            "hostWwn": CONFIG_MAP['host_wwn'],
        },
    ],
}

REMOTE_GET_HOST_WWNS_RESULT = {
    "data": [
        {
            "hostGroupNumber": 0,
            "hostWwn": REMOTE_CONFIG_MAP['host_wwn'],
        },
    ],
}

COMPLETED_SUCCEEDED_RESULT = {
    "status": "Completed",
    "state": "Succeeded",
    "affectedResources": ('a/b/c/1',),
}

REMOTE_COMPLETED_SUCCEEDED_RESULT = {
    "status": "Completed",
    "state": "Succeeded",
    "affectedResources": ('a/b/c/2',),
}

COMPLETED_FAILED_RESULT_LU_DEFINED = {
    "status": "Completed",
    "state": "Failed",
    "error": {
        "errorCode": {
            "SSB1": "B958",
            "SSB2": "015A",
        },
    },
}

GET_LDEV_RESULT = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "HDP"],
    "status": "NML",
    "poolId": 30,
    "dataReductionStatus": "DISABLED",
    "dataReductionMode": "disabled",
    "label": "00000000000000000000000000000000",
}

GET_LDEV_RESULT_SPLIT = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "HDP"],
    "status": "NML",
    "poolId": 30,
    "dataReductionStatus": "DISABLED",
    "dataReductionMode": "disabled",
    "label": "00000000000000000000000000000004",
}

GET_LDEV_RESULT_LABEL = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "HDP"],
    "status": "NML",
    "poolId": 30,
    "dataReductionStatus": "DISABLED",
    "dataReductionMode": "disabled",
    "label": "00000000000000000000000000000001",
}

GET_LDEV_RESULT_MAPPED = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "HDP"],
    "status": "NML",
    "ports": [
        {
            "portId": CONFIG_MAP['port_id'],
            "hostGroupNumber": 0,
            "hostGroupName": CONFIG_MAP['host_grp_name'],
            "lun": 1
        },
    ],
}

REMOTE_GET_LDEV_RESULT_MAPPED = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "HDP"],
    "status": "NML",
    "ports": [
        {
            "portId": REMOTE_CONFIG_MAP['port_id'],
            "hostGroupNumber": 0,
            "hostGroupName": REMOTE_CONFIG_MAP['host_grp_name'],
            "lun": 1
        },
    ],
}

GET_LDEV_RESULT_PAIR = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "HDP", "HTI"],
    "status": "NML",
    "label": "10000000000000000000000000000000",
}

GET_LDEV_RESULT_REP = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "HDP", "GAD"],
    "status": "NML",
    "numOfPorts": 1,
    "label": "00000000000000000000000000000004",
}

GET_LDEV_RESULT_REP_LABEL = {
    "emulationType": "OPEN-V-CVS",
    "blockCapacity": 2097152,
    "attributes": ["CVS", "HDP", "GAD"],
    "status": "NML",
    "numOfPorts": 1,
    "label": "00000000000000000000000000000001",
}

GET_POOL_RESULT = {
    "availableVolumeCapacity": 480144,
    "totalPoolCapacity": 507780,
    "totalLocatedCapacity": 71453172,
    "virtualVolumeCapacityRate": -1,
}

GET_POOLS_RESULT = {
    "data": [
        {
            "poolId": 30,
            "poolName": CONFIG_MAP['pool_name'],
            "availableVolumeCapacity": 480144,
            "totalPoolCapacity": 507780,
            "totalLocatedCapacity": 71453172,
            "virtualVolumeCapacityRate": -1,
        },
    ],
}

GET_SNAPSHOTS_RESULT = {
    "data": [
        {
            "primaryOrSecondary": "S-VOL",
            "status": "PSUS",
            "pvolLdevId": 0,
            "muNumber": 1,
            "svolLdevId": 1,
        },
    ],
}

GET_SNAPSHOTS_RESULT_PAIR = {
    "data": [
        {
            "primaryOrSecondary": "S-VOL",
            "status": "PAIR",
            "pvolLdevId": 0,
            "muNumber": 1,
            "svolLdevId": 1,
        },
    ],
}

GET_SNAPSHOTS_RESULT_BUSY = {
    "data": [
        {
            "primaryOrSecondary": "P-VOL",
            "status": "PSUP",
            "pvolLdevId": 0,
            "muNumber": 1,
            "svolLdevId": 1,
        },
    ],
}

GET_SNAPSHOTS_RESULT_TEST = {
    "data": [
        {
            "primaryOrSecondary": "S-VOL",
            "status": "PSUS",
            "pvolLdevId": 1,
            "muNumber": 1,
            "svolLdevId": 1,
        },
    ],
}

GET_LUNS_RESULT = {
    "data": [
        {
            "ldevId": 0,
            "lun": 1,
        },
    ],
}

GET_HOST_GROUP_RESULT = {
    "hostGroupName": CONFIG_MAP['host_grp_name'],
}

GET_HOST_GROUPS_RESULT = {
    "data": [
        {
            "hostGroupNumber": 0,
            "portId": CONFIG_MAP['port_id'],
            "hostGroupName": "HBSD-test",
        },
    ],
}

GET_HOST_GROUPS_RESULT_PAIR = {
    "data": [
        {
            "hostGroupNumber": 1,
            "portId": CONFIG_MAP['port_id'],
            "hostGroupName": "HBSD-pair00",
        },
    ],
}

REMOTE_GET_HOST_GROUPS_RESULT_PAIR = {
    "data": [
        {
            "hostGroupNumber": 1,
            "portId": REMOTE_CONFIG_MAP['port_id'],
            "hostGroupName": "HBSD-pair00",
        },
    ],
}

GET_LDEVS_RESULT = {
    "data": [
        {
            "ldevId": 0,
            "label": "15960cc738c94c5bb4f1365be5eeed44",
        },
        {
            "ldevId": 1,
            "label": "15960cc738c94c5bb4f1365be5eeed45",
        },
    ],
}

GET_REMOTE_MIRROR_COPYPAIR_RESULT = {
    'pvolLdevId': 4,
    'svolLdevId': 4,
    'pvolStatus': 'PAIR',
    'svolStatus': 'PAIR',
    'replicationType': hbsd_utils.MIRROR_ATTR,
}

GET_REMOTE_MIRROR_COPYPAIR_RESULT_SPLIT = {
    'pvolLdevId': 4,
    'svolLdevId': 4,
    'pvolStatus': 'PSUS',
    'svolStatus': 'SSUS',
    'replicationType': hbsd_utils.MIRROR_ATTR,
}

GET_REMOTE_MIRROR_COPYGROUP_RESULT = {
    'copyGroupName': 'HBSD-127.0.0.100G00',
    'copyPairs': [GET_REMOTE_MIRROR_COPYPAIR_RESULT],
}

GET_REMOTE_MIRROR_COPYGROUP_RESULT_ERROR = {
    "errorSource": "<URL>",
    "message": "<message>",
    "solution": "<solution>",
    "messageId": "aaa",
    "errorCode": {
                   "SSB1": "",
                   "SSB2": "",
    }
}

NOTFOUND_RESULT = {
    "data": [],
}

ERROR_RESULT = {
    "errorSource": "<URL>",
    "message": "<message>",
    "solution": "<solution>",
    "messageId": "<messageId>",
    "errorCode": {
                   "SSB1": "",
                   "SSB2": "",
    }
}


def _brick_get_connector_properties(multipath=False, enforce_multipath=False):
    """Return a predefined connector object."""
    return DEFAULT_CONNECTOR


class FakeLookupService():
    """Dummy FC zoning mapping lookup service class."""

    def get_device_mapping_from_network(self, initiator_wwns, target_wwns):
        """Return predefined FC zoning mapping."""
        return DEVICE_MAP


class FakeResponse():

    def __init__(self, status_code, data=None, headers=None):
        self.status_code = status_code
        self.data = data
        self.text = data
        self.content = data
        self.headers = {'Content-Type': 'json'} if headers is None else headers

    def json(self):
        return self.data


class HBSDMIRRORFCDriverTest(test.TestCase):
    """Unit test class for HBSD MIRROR interface fibre channel module."""

    test_existing_ref = {'source-id': '1'}
    test_existing_ref_name = {
        'source-name': '15960cc7-38c9-4c5b-b4f1-365be5eeed45'}

    def setUp(self):
        """Set up the test environment."""
        def _set_required(opts, required):
            for opt in opts:
                opt.required = required

        # Initialize Cinder and avoid checking driver options.
        rest_required_opts = [
            opt for opt in hbsd_rest.REST_VOLUME_OPTS if opt.required]
        common_required_opts = [
            opt for opt in hbsd_common.COMMON_VOLUME_OPTS if opt.required]
        _set_required(rest_required_opts, False)
        _set_required(common_required_opts, False)
        super(HBSDMIRRORFCDriverTest, self).setUp()
        _set_required(rest_required_opts, True)
        _set_required(common_required_opts, True)

        self.configuration = conf.Configuration(None)
        self.ctxt = cinder_context.get_admin_context()
        self._setup_config()
        self._setup_driver()

    def _setup_config(self):
        """Set configuration parameter values."""
        self.configuration.config_group = "REST"

        self.configuration.volume_backend_name = "RESTFC"
        self.configuration.volume_driver = (
            "cinder.volume.drivers.hitachi.hbsd_fc.HBSDFCDriver")
        self.configuration.reserved_percentage = "0"
        self.configuration.use_multipath_for_image_xfer = False
        self.configuration.enforce_multipath_for_image_xfer = False
        self.configuration.max_over_subscription_ratio = 500.0
        self.configuration.driver_ssl_cert_verify = False

        self.configuration.hitachi_storage_id = CONFIG_MAP['serial']
        self.configuration.hitachi_pools = ["30"]
        self.configuration.hitachi_snap_pool = None
        self.configuration.hitachi_ldev_range = "0-1"
        self.configuration.hitachi_target_ports = [CONFIG_MAP['port_id']]
        self.configuration.hitachi_compute_target_ports\
            = [CONFIG_MAP['port_id']]
        self.configuration.hitachi_group_create = True
        self.configuration.hitachi_group_delete = True
        self.configuration.hitachi_copy_speed = 3
        self.configuration.hitachi_copy_check_interval = 3
        self.configuration.hitachi_async_copy_check_interval = 10

        self.configuration.san_login = CONFIG_MAP['user_id']
        self.configuration.san_password = CONFIG_MAP['user_pass']
        self.configuration.san_ip = CONFIG_MAP[
            'rest_server_ip_addr']
        self.configuration.san_api_port = CONFIG_MAP[
            'rest_server_ip_port']
        self.configuration.hitachi_rest_disable_io_wait = True
        self.configuration.hitachi_rest_tcp_keepalive = True
        self.configuration.hitachi_discard_zero_page = True
        self.configuration.hitachi_lun_timeout = hbsd_rest._LUN_TIMEOUT
        self.configuration.hitachi_lun_retry_interval = (
            hbsd_rest._LUN_RETRY_INTERVAL)
        self.configuration.hitachi_restore_timeout = hbsd_rest._RESTORE_TIMEOUT
        self.configuration.hitachi_state_transition_timeout = (
            hbsd_rest._STATE_TRANSITION_TIMEOUT)
        self.configuration.hitachi_lock_timeout = hbsd_rest_api._LOCK_TIMEOUT
        self.configuration.hitachi_rest_timeout = hbsd_rest_api._REST_TIMEOUT
        self.configuration.hitachi_extend_timeout = (
            hbsd_rest_api._EXTEND_TIMEOUT)
        self.configuration.hitachi_exec_retry_interval = (
            hbsd_rest_api._EXEC_RETRY_INTERVAL)
        self.configuration.hitachi_rest_connect_timeout = (
            hbsd_rest_api._DEFAULT_CONNECT_TIMEOUT)
        self.configuration.hitachi_rest_job_api_response_timeout = (
            hbsd_rest_api._JOB_API_RESPONSE_TIMEOUT)
        self.configuration.hitachi_rest_get_api_response_timeout = (
            hbsd_rest_api._GET_API_RESPONSE_TIMEOUT)
        self.configuration.hitachi_rest_server_busy_timeout = (
            hbsd_rest_api._REST_SERVER_BUSY_TIMEOUT)
        self.configuration.hitachi_rest_keep_session_loop_interval = (
            hbsd_rest_api._KEEP_SESSION_LOOP_INTERVAL)
        self.configuration.hitachi_rest_another_ldev_mapped_retry_timeout = (
            hbsd_rest_api._ANOTHER_LDEV_MAPPED_RETRY_TIMEOUT)
        self.configuration.hitachi_rest_tcp_keepidle = (
            hbsd_rest_api._TCP_KEEPIDLE)
        self.configuration.hitachi_rest_tcp_keepintvl = (
            hbsd_rest_api._TCP_KEEPINTVL)
        self.configuration.hitachi_rest_tcp_keepcnt = (
            hbsd_rest_api._TCP_KEEPCNT)
        self.configuration.hitachi_host_mode_options = []

        self.configuration.hitachi_zoning_request = False

        self.configuration.use_chap_auth = True
        self.configuration.chap_username = CONFIG_MAP['auth_user']
        self.configuration.chap_password = CONFIG_MAP['auth_password']

        self.configuration.san_thin_provision = True
        self.configuration.san_private_key = ''
        self.configuration.san_clustername = ''
        self.configuration.san_ssh_port = '22'
        self.configuration.san_is_local = False
        self.configuration.ssh_conn_timeout = '30'
        self.configuration.ssh_min_pool_conn = '1'
        self.configuration.ssh_max_pool_conn = '5'

        self.configuration.hitachi_replication_status_check_short_interval = 5
        self.configuration.hitachi_replication_status_check_long_interval\
            = 10 * 60
        self.configuration.hitachi_replication_status_check_timeout\
            = 24 * 60 * 60

        self.configuration.hitachi_replication_number = 0
        self.configuration.hitachi_pair_target_number = 0
        self.configuration.hitachi_rest_pair_target_ports\
            = [CONFIG_MAP['port_id']]
        self.configuration.hitachi_quorum_disk_id = 13
        self.configuration.hitachi_mirror_copy_speed = 3
        self.configuration.hitachi_mirror_storage_id\
            = REMOTE_CONFIG_MAP['serial']
        self.configuration.hitachi_mirror_pool = '40'
        self.configuration.hitachi_mirror_snap_pool = None
        self.configuration.hitachi_mirror_ldev_range = '2-3'
        self.configuration.hitachi_mirror_target_ports\
            = [REMOTE_CONFIG_MAP['port_id']]
        self.configuration.hitachi_mirror_compute_target_ports\
            = [REMOTE_CONFIG_MAP['port_id']]
        self.configuration.hitachi_mirror_pair_target_number = 0
        self.configuration.hitachi_mirror_rest_pair_target_ports\
            = [REMOTE_CONFIG_MAP['port_id']]
        self.configuration.hitachi_mirror_rest_user\
            = REMOTE_CONFIG_MAP['user_id']
        self.configuration.hitachi_mirror_rest_password\
            = REMOTE_CONFIG_MAP['user_pass']
        self.configuration.hitachi_mirror_rest_api_ip\
            = REMOTE_CONFIG_MAP['rest_server_ip_addr']
        self.configuration.hitachi_mirror_rest_api_port\
            = REMOTE_CONFIG_MAP['rest_server_ip_port']
        self.configuration.hitachi_set_mirror_reserve_attribute = True
        self.configuration.hitachi_path_group_id = 0

        self.configuration.hitachi_mirror_use_chap_auth = True
        self.configuration.hitachi_mirror_chap_user = CONFIG_MAP['auth_user']
        self.configuration.hitachi_mirror_chap_password\
            = CONFIG_MAP['auth_password']

        self.configuration.hitachi_mirror_ssl_cert_verify = False
        self.configuration.hitachi_mirror_ssl_cert_path = '/root/path'

        self.configuration.safe_get = self._fake_safe_get

        CONF = cfg.CONF
        CONF.my_ip = CONFIG_MAP['my_ip']

    def _fake_safe_get(self, value):
        """Retrieve a configuration value avoiding throwing an exception."""
        try:
            val = getattr(self.configuration, value)
        except AttributeError:
            val = None
        return val

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(
        volume_utils, 'brick_get_connector_properties',
        side_effect=_brick_get_connector_properties)
    def _setup_driver(
            self, brick_get_connector_properties=None, request=None):
        """Set up the driver environment."""
        self.driver = hbsd_fc.HBSDFCDriver(
            configuration=self.configuration)

        def _request_side_effect(
                method, url, params, json, headers, auth, timeout, verify):
            if self.configuration.hitachi_storage_id in url:
                if method == 'POST':
                    return FakeResponse(200, POST_SESSIONS_RESULT)
                elif '/ports' in url:
                    return FakeResponse(200, GET_PORTS_RESULT)
                elif '/host-wwns' in url:
                    return FakeResponse(200, GET_HOST_WWNS_RESULT)
                elif '/host-groups' in url:
                    return FakeResponse(200, GET_HOST_GROUPS_RESULT_PAIR)
            else:
                if method == 'POST':
                    return FakeResponse(200, REMOTE_POST_SESSIONS_RESULT)
                elif '/ports' in url:
                    return FakeResponse(200, REMOTE_GET_PORTS_RESULT)
                elif '/host-wwns' in url:
                    return FakeResponse(200, REMOTE_GET_HOST_WWNS_RESULT)
                elif '/host-groups' in url:
                    return FakeResponse(
                        200, REMOTE_GET_HOST_GROUPS_RESULT_PAIR)
            return FakeResponse(
                500, ERROR_RESULT, headers={'Content-Type': 'json'})
        request.side_effect = _request_side_effect
        self.driver.do_setup(None)
        self.driver.check_for_setup_error()
        self.driver.local_path(None)
        self.driver.create_export(None, None, None)
        self.driver.ensure_export(None, None)
        self.driver.remove_export(None, None)
        self.driver.create_export_snapshot(None, None, None)
        self.driver.remove_export_snapshot(None, None)
        # stop the Loopingcall within the do_setup treatment
        self.driver.common.rep_primary.client.keep_session_loop.stop()
        self.driver.common.rep_secondary.client.keep_session_loop.stop()

    def tearDown(self):
        self.client = None
        super(HBSDMIRRORFCDriverTest, self).tearDown()

    # API test cases
    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(
        volume_utils, 'brick_get_connector_properties',
        side_effect=_brick_get_connector_properties)
    def test_do_setup(self, brick_get_connector_properties, request):
        drv = hbsd_fc.HBSDFCDriver(
            configuration=self.configuration)
        self._setup_config()
        self.configuration.hitachi_pair_target_number = 10
        self.configuration.hitachi_mirror_pair_target_number = 20

        def _request_side_effect(
                method, url, params, json, headers, auth, timeout, verify):
            if self.configuration.hitachi_storage_id in url:
                if method == 'POST':
                    return FakeResponse(200, POST_SESSIONS_RESULT)
                elif '/ports' in url:
                    return FakeResponse(200, GET_PORTS_RESULT)
                elif '/host-wwns' in url:
                    return FakeResponse(200, GET_HOST_WWNS_RESULT)
                elif '/host-groups' in url:
                    return FakeResponse(200, GET_HOST_GROUPS_RESULT_PAIR)
            else:
                if method == 'POST':
                    return FakeResponse(200, REMOTE_POST_SESSIONS_RESULT)
                elif '/ports' in url:
                    return FakeResponse(200, REMOTE_GET_PORTS_RESULT)
                elif '/host-wwns' in url:
                    return FakeResponse(200, REMOTE_GET_HOST_WWNS_RESULT)
                elif '/host-groups' in url:
                    return FakeResponse(
                        200, REMOTE_GET_HOST_GROUPS_RESULT_PAIR)
            return FakeResponse(
                500, ERROR_RESULT, headers={'Content-Type': 'json'})
        request.side_effect = _request_side_effect
        drv.do_setup(None)
        self.assertEqual(
            {CONFIG_MAP['port_id']: CONFIG_MAP['target_wwn']},
            drv.common.rep_primary.storage_info['wwns'])
        self.assertEqual(
            {REMOTE_CONFIG_MAP['port_id']: REMOTE_CONFIG_MAP['target_wwn']},
            drv.common.rep_secondary.storage_info['wwns'])
        self.assertEqual(2, brick_get_connector_properties.call_count)
        self.assertEqual(10, request.call_count)
        self.assertEqual(
            "HBSD-pair%2d" % self.configuration.hitachi_pair_target_number,
            drv.common.rep_primary._PAIR_TARGET_NAME)
        self.assertEqual(
            ("HBSD-pair%2d" %
             self.configuration.hitachi_mirror_pair_target_number),
            drv.common.rep_secondary._PAIR_TARGET_NAME)
        # stop the Loopingcall within the do_setup treatment
        drv.common.rep_primary.client.keep_session_loop.stop()
        drv.common.rep_secondary.client.keep_session_loop.stop()
        self._setup_config()

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_volume(self, get_volume_type_qos_specs,
                           get_volume_type_extra_specs, request):
        extra_specs = {"test1": "aaa"}
        get_volume_type_extra_specs.return_value = extra_specs
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.return_value = FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        ret = self.driver.create_volume(TEST_VOLUME[7])
        actual = {'provider_location': '1'}
        self.assertEqual(actual, ret)
        self.assertEqual(2, request.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_volume_replication(self, get_volume_type_qos_specs,
                                       get_volume_type_extra_specs, request):
        extra_specs = {"test1": "aaa",
                       "hbsd:topology": "active_active_mirror_volume"}
        get_volume_type_extra_specs.return_value = extra_specs
        get_volume_type_qos_specs.return_value = {'qos_specs': None}

        def _request_side_effect(
                method, url, params, json, headers, auth, timeout, verify):
            if self.configuration.hitachi_storage_id in url:
                if method in ('POST', 'PUT'):
                    return FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/remote-mirror-copygroups' in url:
                        return FakeResponse(200, NOTFOUND_RESULT)
                    elif '/remote-mirror-copypairs/' in url:
                        return FakeResponse(
                            200, GET_REMOTE_MIRROR_COPYPAIR_RESULT)
            else:
                if method in ('POST', 'PUT'):
                    return FakeResponse(202, REMOTE_COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/remote-mirror-copygroups' in url:
                        return FakeResponse(200, NOTFOUND_RESULT)
            return FakeResponse(
                500, ERROR_RESULT, headers={'Content-Type': 'json'})
        request.side_effect = _request_side_effect
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        ret = self.driver.create_volume(TEST_VOLUME[3])
        actual = {
            'provider_location': json.dumps(
                {'pldev': 1, 'sldev': 2,
                 'remote-copy': hbsd_utils.MIRROR_ATTR})}
        self.assertEqual(actual, ret)
        self.assertEqual(14, request.call_count)
        for args, kwargs in request.call_args_list:
            if args[0] == 'POST' and 'remote-mirror-copypairs' in args[1]:
                self.assertEqual('G', kwargs['json']['copyGroupName'][-3])
                break
        else:
            self.fail('no create pair api')

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_volume_replication_qos(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        input_qos_specs = {
            'qos_specs': {
                'consumer': 'back-end',
                'specs': {'upperIops': '1000'}}}
        get_volume_type_qos_specs.return_value = input_qos_specs
        extra_specs = {"test1": "aaa",
                       "hbsd:topology": "active_active_mirror_volume"}
        get_volume_type_extra_specs.return_value = extra_specs

        def _request_side_effect(
                method, url, params, json, headers, auth, timeout, verify):
            if self.configuration.hitachi_storage_id in url:
                if method in ('POST', 'PUT'):
                    return FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/remote-mirror-copygroups' in url:
                        return FakeResponse(200, NOTFOUND_RESULT)
                    elif '/remote-mirror-copypairs/' in url:
                        return FakeResponse(
                            200, GET_REMOTE_MIRROR_COPYPAIR_RESULT)
            else:
                if method in ('POST', 'PUT'):
                    return FakeResponse(202, REMOTE_COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/remote-mirror-copygroups' in url:
                        return FakeResponse(200, NOTFOUND_RESULT)
            return FakeResponse(
                500, ERROR_RESULT, headers={'Content-Type': 'json'})
        request.side_effect = _request_side_effect
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        ret = self.driver.create_volume(TEST_VOLUME[3])
        actual = {
            'provider_location': json.dumps(
                {'pldev': 1, 'sldev': 2,
                 'remote-copy': hbsd_utils.MIRROR_ATTR})}
        self.assertEqual(actual, ret)
        self.assertEqual(1, get_volume_type_extra_specs.call_count)
        self.assertEqual(1, get_volume_type_qos_specs.call_count)
        self.assertEqual(16, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_volume(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.delete_volume(TEST_VOLUME[0])
        self.assertEqual(5, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_volume_replication(self, request):
        self.copygroup_count = 0
        self.ldev_count = 0

        def _request_side_effect(
                method, url, params, json, headers, auth, timeout, verify):
            if self.configuration.hitachi_storage_id in url:
                if method in ('POST', 'PUT', 'DELETE'):
                    return FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/remote-mirror-copygroups/' in url:
                        if self.copygroup_count < 2:
                            self.copygroup_count = self.copygroup_count + 1
                            return FakeResponse(
                                200, GET_REMOTE_MIRROR_COPYGROUP_RESULT)
                        else:
                            return FakeResponse(
                                500, GET_REMOTE_MIRROR_COPYGROUP_RESULT_ERROR,
                                headers={'Content-Type': 'json'})
                    elif '/remote-mirror-copypairs/' in url:
                        return FakeResponse(
                            200, GET_REMOTE_MIRROR_COPYPAIR_RESULT_SPLIT)
                    elif '/ldevs/' in url:
                        if self.ldev_count == 0:
                            self.ldev_count = self.ldev_count + 1
                            return FakeResponse(200, GET_LDEV_RESULT_REP)
                        else:
                            return FakeResponse(200, GET_LDEV_RESULT_SPLIT)
            else:
                if method in ('POST', 'PUT', 'DELETE'):
                    return FakeResponse(202, REMOTE_COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/ldevs/' in url:
                        return FakeResponse(200, GET_LDEV_RESULT_SPLIT)
            return FakeResponse(
                500, ERROR_RESULT, headers={'Content-Type': 'json'})
        request.side_effect = _request_side_effect
        self.driver.delete_volume(TEST_VOLUME[4])
        self.assertEqual(17, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_volume_primary_is_invalid_ldev(self, request):
        request.return_value = FakeResponse(200, GET_LDEV_RESULT_LABEL)
        self.driver.delete_volume(TEST_VOLUME[0])
        self.assertEqual(1, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_volume_primary_secondary_is_invalid_ldev(self, request):
        request.return_value = FakeResponse(200, GET_LDEV_RESULT_REP_LABEL)
        self.driver.delete_volume(TEST_VOLUME[4])
        self.assertEqual(2, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_volume_secondary_is_invalid_ldev(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_REP_LABEL),
                               FakeResponse(200, GET_LDEV_RESULT_REP),
                               FakeResponse(200, GET_LDEV_RESULT_REP),
                               FakeResponse(200, GET_LDEV_RESULT_REP),
                               FakeResponse(200, GET_LDEV_RESULT_REP),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.delete_volume(TEST_VOLUME[4])
        self.assertEqual(6, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_extend_volume(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.extend_volume(TEST_VOLUME[0], 256)
        self.assertEqual(4, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_extend_volume_replication(self, request):
        self.ldev_count = 0
        self.copypair_count = 0

        def _request_side_effect(
                method, url, params, json, headers, auth, timeout, verify):
            if self.configuration.hitachi_storage_id in url:
                if method in ('POST', 'PUT', 'DELETE'):
                    return FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/remote-mirror-copygroups/' in url:
                        return FakeResponse(
                            200, GET_REMOTE_MIRROR_COPYGROUP_RESULT)
                    elif '/remote-mirror-copygroups' in url:
                        return FakeResponse(200, NOTFOUND_RESULT)
                    elif '/remote-mirror-copypairs/' in url:
                        if self.copypair_count == 0:
                            self.copypair_count = self.copypair_count + 1
                            return FakeResponse(
                                200, GET_REMOTE_MIRROR_COPYPAIR_RESULT_SPLIT)
                        else:
                            return FakeResponse(
                                200, GET_REMOTE_MIRROR_COPYPAIR_RESULT)
                    elif '/ldevs/' in url:
                        if self.ldev_count < 2:
                            self.ldev_count = self.ldev_count + 1
                            return FakeResponse(200, GET_LDEV_RESULT_REP)
                        else:
                            return FakeResponse(200, GET_LDEV_RESULT)
            else:
                if method in ('POST', 'PUT', 'DELETE'):
                    return FakeResponse(202, REMOTE_COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/ldevs/' in url:
                        return FakeResponse(200, GET_LDEV_RESULT)
            return FakeResponse(
                500, ERROR_RESULT, headers={'Content-Type': 'json'})
        request.side_effect = _request_side_effect
        self.driver.extend_volume(TEST_VOLUME[4], 256)
        self.assertEqual(23, request.call_count)

    @mock.patch.object(driver.FibreChannelDriver, "get_goodness_function")
    @mock.patch.object(driver.FibreChannelDriver, "get_filter_function")
    @mock.patch.object(requests.Session, "request")
    def test_get_volume_stats(
            self, request, get_filter_function, get_goodness_function):
        request.return_value = FakeResponse(200, GET_POOLS_RESULT)
        get_filter_function.return_value = None
        get_goodness_function.return_value = None
        stats = self.driver.get_volume_stats(True)
        self.assertEqual('Hitachi', stats['vendor_name'])
        self.assertTrue(stats["pools"][0]['multiattach'])
        self.assertEqual(1, request.call_count)
        self.assertEqual(1, get_filter_function.call_count)
        self.assertEqual(1, get_goodness_function.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(sqlalchemy_api, 'volume_get', side_effect=_volume_get)
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_snapshot(
            self, get_volume_type_qos_specs, volume_get,
            get_volume_type_extra_specs, request):
        extra_specs = {"test1": "aaa",
                       "hbsd:topology": "active_active_mirror_volume"}
        get_volume_type_extra_specs.return_value = extra_specs
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        ret = self.driver.create_snapshot(TEST_SNAPSHOT[0])
        actual = {'provider_location': '1'}
        self.assertEqual(actual, ret)
        self.assertEqual(5, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_snapshot(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.delete_snapshot(TEST_SNAPSHOT[0])
        self.assertEqual(14, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_snapshot_pldev_in_loc(self, request):
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.delete_snapshot,
                          TEST_SNAPSHOT[1])
        self.assertEqual(1, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_delete_snapshot_snapshot_is_busy(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT_TEST)]
        self.assertRaises(exception.SnapshotIsBusy,
                          self.driver.delete_snapshot,
                          TEST_SNAPSHOT[0])
        self.assertEqual(3, request.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_cloned_volume(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        extra_specs = {"test1": "aaa"}
        get_volume_type_extra_specs.return_value = extra_specs
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        ret = self.driver.create_cloned_volume(TEST_VOLUME[0], TEST_VOLUME[1])
        actual = {'provider_location': '1'}
        self.assertEqual(actual, ret)
        self.assertEqual(5, request.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_cloned_volume_replication(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        extra_specs = {"hbsd:topology": "active_active_mirror_volume"}
        get_volume_type_extra_specs.return_value = extra_specs
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        self.snapshot_count = 0

        def _request_side_effect(
                method, url, params, json, headers, auth, timeout, verify):
            if self.configuration.hitachi_storage_id in url:
                if method in ('POST', 'PUT'):
                    return FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/remote-mirror-copygroups' in url:
                        return FakeResponse(200, NOTFOUND_RESULT)
                    elif '/remote-mirror-copypairs/' in url:
                        return FakeResponse(
                            200, GET_REMOTE_MIRROR_COPYPAIR_RESULT)
                    elif '/ldevs/' in url:
                        return FakeResponse(200, GET_LDEV_RESULT_REP)
                    elif '/snapshots' in url:
                        if self.snapshot_count < 1:
                            self.snapshot_count = self.snapshot_count + 1
                            return FakeResponse(200, GET_SNAPSHOTS_RESULT)
                        else:
                            return FakeResponse(200, NOTFOUND_RESULT)
            else:
                if method in ('POST', 'PUT'):
                    return FakeResponse(202, REMOTE_COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/remote-mirror-copygroups' in url:
                        return FakeResponse(200, NOTFOUND_RESULT)
            return FakeResponse(
                500, ERROR_RESULT, headers={'Content-Type': 'json'})
        request.side_effect = _request_side_effect
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        ret = self.driver.create_cloned_volume(TEST_VOLUME[4], TEST_VOLUME[5])
        actual = {
            'provider_location': json.dumps(
                {'pldev': 1, 'sldev': 2,
                 'remote-copy': hbsd_utils.MIRROR_ATTR})}
        self.assertEqual(actual, ret)
        self.assertEqual(23, request.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_volume_from_snapshot(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        extra_specs = {"test1": "aaa"}
        get_volume_type_extra_specs.return_value = extra_specs
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        ret = self.driver.create_volume_from_snapshot(
            TEST_VOLUME[0], TEST_SNAPSHOT[0])
        actual = {'provider_location': '1'}
        self.assertEqual(actual, ret)
        self.assertEqual(5, request.call_count)

    @mock.patch.object(fczm_utils, "add_fc_zone")
    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    def test_initialize_connection(
            self, get_volume_type_extra_specs, request, add_fc_zone):
        self.driver.common.conf.hitachi_zoning_request = True
        self.driver.common.rep_primary.lookup_service = FakeLookupService()
        self.driver.common.rep_secondary.lookup_service = FakeLookupService()
        extra_specs = {"test1": "aaa"}
        get_volume_type_extra_specs.return_value = extra_specs
        def _request_side_effect(
                method, url, params, json, headers, auth, timeout, verify):
            if self.configuration.hitachi_storage_id in url:
                if method in ('POST', 'PUT'):
                    return FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    return FakeResponse(200, GET_HOST_WWNS_RESULT)
            else:
                if method in ('POST', 'PUT'):
                    return FakeResponse(202, REMOTE_COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    return FakeResponse(200, REMOTE_GET_HOST_WWNS_RESULT)
            return FakeResponse(
                500, ERROR_RESULT, headers={'Content-Type': 'json'})
        request.side_effect = _request_side_effect
        ret = self.driver.initialize_connection(
            TEST_VOLUME[4], DEFAULT_CONNECTOR)
        self.assertEqual('fibre_channel', ret['driver_volume_type'])
        self.assertEqual(
            [CONFIG_MAP['target_wwn'], REMOTE_CONFIG_MAP['target_wwn']],
            ret['data']['target_wwn'])
        self.assertEqual(1, ret['data']['target_lun'])
        self.assertEqual(4, request.call_count)
        self.assertEqual(1, add_fc_zone.call_count)

    @mock.patch.object(fczm_utils, "remove_fc_zone")
    @mock.patch.object(requests.Session, "request")
    def test_terminate_connection(self, request, remove_fc_zone):
        self.driver.common.conf.hitachi_zoning_request = True
        self.driver.common.rep_primary.lookup_service = FakeLookupService()
        self.driver.common.rep_secondary.lookup_service = FakeLookupService()
        def _request_side_effect(
                method, url, params, json, headers, auth, timeout, verify):
            if self.configuration.hitachi_storage_id in url:
                if method in ('POST', 'PUT', 'DELETE'):
                    return FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/ldevs/' in url:
                        return FakeResponse(200, GET_LDEV_RESULT_MAPPED)
                    elif '/host-wwns' in url:
                        return FakeResponse(200, GET_HOST_WWNS_RESULT)
                    else:
                        return FakeResponse(200, NOTFOUND_RESULT)
            else:
                if method in ('POST', 'PUT', 'DELETE'):
                    return FakeResponse(202, REMOTE_COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/ldevs/' in url:
                        return FakeResponse(200, REMOTE_GET_LDEV_RESULT_MAPPED)
                    elif '/host-wwns' in url:
                        return FakeResponse(200, REMOTE_GET_HOST_WWNS_RESULT)
                    else:
                        return FakeResponse(200, NOTFOUND_RESULT)
            return FakeResponse(
                500, ERROR_RESULT, headers={'Content-Type': 'json'})
        request.side_effect = _request_side_effect
        self.driver.terminate_connection(TEST_VOLUME[6], DEFAULT_CONNECTOR)
        self.assertEqual(10, request.call_count)
        self.assertEqual(1, remove_fc_zone.call_count)

    @mock.patch.object(fczm_utils, "add_fc_zone")
    @mock.patch.object(requests.Session, "request")
    def test_initialize_connection_snapshot(self, request, add_fc_zone):
        self.driver.common.rep_primary.conf.hitachi_zoning_request = True
        self.driver.common.lookup_service = FakeLookupService()
        request.side_effect = [FakeResponse(200, GET_HOST_WWNS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        ret = self.driver.initialize_connection_snapshot(
            TEST_SNAPSHOT[0], DEFAULT_CONNECTOR)
        self.assertEqual('fibre_channel', ret['driver_volume_type'])
        self.assertEqual([CONFIG_MAP['target_wwn']], ret['data']['target_wwn'])
        self.assertEqual(1, ret['data']['target_lun'])
        self.assertEqual(2, request.call_count)
        self.assertEqual(1, add_fc_zone.call_count)

    @mock.patch.object(fczm_utils, "remove_fc_zone")
    @mock.patch.object(requests.Session, "request")
    def test_terminate_connection_snapshot(self, request, remove_fc_zone):
        self.driver.common.rep_primary.conf.hitachi_zoning_request = True
        self.driver.common.lookup_service = FakeLookupService()
        request.side_effect = [FakeResponse(200, GET_HOST_WWNS_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT_MAPPED),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.terminate_connection_snapshot(
            TEST_SNAPSHOT[0], DEFAULT_CONNECTOR)
        self.assertEqual(5, request.call_count)
        self.assertEqual(1, remove_fc_zone.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_manage_existing(self, get_volume_type_qos_specs, request):
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_LDEVS_RESULT)]
        ret = self.driver.manage_existing(
            TEST_VOLUME[0], self.test_existing_ref)
        actual = {'provider_location': '1'}
        self.assertEqual(actual, ret)
        self.assertEqual(3, request.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    def test_manage_existing_get_size(
            self, get_volume_type_extra_specs, request):
        extra_specs = {"test1": "aaa"}
        get_volume_type_extra_specs.return_value = extra_specs
        request.return_value = FakeResponse(200, GET_LDEV_RESULT)
        self.driver.manage_existing_get_size(
            TEST_VOLUME[0], self.test_existing_ref)
        self.assertEqual(1, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_unmanage(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT)]
        self.driver.unmanage(TEST_VOLUME[0])
        self.assertEqual(3, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_unmanage_has_rep_pair_true(self, request):
        request.return_value = FakeResponse(200, GET_LDEV_RESULT_REP)
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.unmanage,
                          TEST_VOLUME[4])
        self.assertEqual(1, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_copy_image_to_volume(self, request):
        image_service = 'fake_image_service'
        image_id = 'fake_image_id'
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, COMPLETED_SUCCEEDED_RESULT)]
        with mock.patch.object(driver.VolumeDriver, 'copy_image_to_volume') \
                as mock_copy_image:
            self.driver.copy_image_to_volume(
                self.ctxt, TEST_VOLUME[0], image_service, image_id)
        mock_copy_image.assert_called_with(
            self.ctxt, TEST_VOLUME[0], image_service, image_id,
            disable_sparse=False)
        self.assertEqual(2, request.call_count)

    @mock.patch.object(requests.Session, "request")
    def test_update_migrated_volume(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        ret = self.driver.update_migrated_volume(
            self.ctxt, TEST_VOLUME[0], TEST_VOLUME[1], "available")
        self.assertEqual(2, request.call_count)
        actual = ({'_name_id': TEST_VOLUME[1]['id'],
                   'provider_location': TEST_VOLUME[1]['provider_location']})
        self.assertEqual(actual, ret)

    @mock.patch.object(requests.Session, "request")
    def test_update_migrated_volume_replication(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_REP),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        ret = self.driver.update_migrated_volume(
            self.ctxt, TEST_VOLUME[0], TEST_VOLUME[4], "available")
        self.assertEqual(3, request.call_count)
        actual = ({'_name_id': TEST_VOLUME[4]['id'],
                   'provider_location': TEST_VOLUME[4]['provider_location']})
        self.assertEqual(actual, ret)

    def test_unmanage_snapshot(self):
        """The driver don't support unmange_snapshot."""
        self.assertRaises(
            NotImplementedError,
            self.driver.unmanage_snapshot,
            TEST_SNAPSHOT[0])

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(obj_snap.SnapshotList, 'get_all_for_volume')
    def test_retype(self, get_all_for_volume,
                    get_volume_type_extra_specs, request):
        extra_specs = {'test1': 'aaa',
                       'hbsd:target_ports': 'CL2-A'}
        get_volume_type_extra_specs.return_value = extra_specs
        get_all_for_volume.return_value = True

        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT)]

        old_specs = {'hbsd:target_ports': 'CL1-A'}
        new_specs = {'hbsd:target_ports': 'CL2-A'}
        old_type_ref = volume_types.create(self.ctxt, 'old', old_specs)
        new_type_ref = volume_types.create(self.ctxt, 'new', new_specs)
        new_type = volume_types.get_volume_type(self.ctxt, new_type_ref['id'])

        diff = volume_types.volume_types_diff(self.ctxt, old_type_ref['id'],
                                              new_type_ref['id'])[0]
        host = {
            'capabilities': {
                'location_info': {
                    'pool_id': 30,
                },
            },
        }

        ret = self.driver.retype(
            self.ctxt, TEST_VOLUME[0], new_type, diff, host)
        self.assertEqual(2, request.call_count)
        self.assertFalse(ret)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    def test_retype_replication(self, get_volume_type_extra_specs, request):
        extra_specs = {'test1': 'aaa',
                       'hbsd:topology': 'active_active_mirror_volume'}
        get_volume_type_extra_specs.return_value = extra_specs

        request.return_value = FakeResponse(200, GET_LDEV_RESULT)

        new_type_ref = volume_types.create(self.ctxt, 'new', extra_specs)
        new_type = volume_types.get_volume_type(self.ctxt, new_type_ref['id'])
        diff = {}
        host = {
            'capabilities': {
                'location_info': {
                    'pool_id': 30,
                },
            },
        }
        ret = self.driver.retype(
            self.ctxt, TEST_VOLUME[0], new_type, diff, host)
        self.assertEqual(1, request.call_count)
        self.assertFalse(ret)

    @mock.patch.object(requests.Session, "request")
    def test_migrate_volume(
            self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT)]
        host = {
            'capabilities': {
                'location_info': {
                    'storage_id': CONFIG_MAP['serial'],
                    'pool_id': 30,
                },
            },
        }
        ret = self.driver.migrate_volume(self.ctxt, TEST_VOLUME[0], host)
        self.assertEqual(3, request.call_count)
        actual = (True, None)
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(requests.Session, "request")
    def test_revert_to_snapshot(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT)]
        self.driver.revert_to_snapshot(
            self.ctxt, TEST_VOLUME[0], TEST_SNAPSHOT[0])
        self.assertEqual(8, request.call_count)

    def test_create_group(self):
        ret = self.driver.create_group(self.ctxt, TEST_GROUP[0])
        self.assertIsNone(ret)

    @mock.patch.object(requests.Session, "request")
    def test_delete_group(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        ret = self.driver.delete_group(
            self.ctxt, TEST_GROUP[0], [TEST_VOLUME[0]])
        self.assertEqual(5, request.call_count)
        actual = (
            {'status': TEST_GROUP[0]['status']},
            [{'id': TEST_VOLUME[0]['id'], 'status': 'deleted'}]
        )
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_group_from_src_volume(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        extra_specs = {"test1": "aaa"}
        get_volume_type_extra_specs.return_value = extra_specs
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        ret = self.driver.create_group_from_src(
            self.ctxt, TEST_GROUP[1], [TEST_VOLUME[1]],
            source_group=TEST_GROUP[0], source_vols=[TEST_VOLUME[0]]
        )
        self.assertEqual(5, request.call_count)
        actual = (
            None,
            [{'id': TEST_VOLUME[1]['id'],
              'provider_location': '1'}])
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_group_from_src_Exception(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        extra_specs = {"test1": "aaa"}
        get_volume_type_extra_specs.return_value = extra_specs
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_group_from_src,
                          self.ctxt, TEST_GROUP[1],
                          [TEST_VOLUME[1], TEST_VOLUME[1]],
                          source_group=TEST_GROUP[0],
                          source_vols=[TEST_VOLUME[0], TEST_VOLUME[3]]
                          )
        self.assertEqual(10, request.call_count)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_group_from_src_snapshot(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        extra_specs = {"test1": "aaa"}
        get_volume_type_extra_specs.return_value = extra_specs
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        ret = self.driver.create_group_from_src(
            self.ctxt, TEST_GROUP[0], [TEST_VOLUME[0]],
            group_snapshot=TEST_GROUP_SNAP[0], snapshots=[TEST_SNAPSHOT[0]]
        )
        self.assertEqual(5, request.call_count)
        actual = (
            None,
            [{'id': TEST_VOLUME[0]['id'],
              'provider_location': '1'}])
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(volume_utils, 'is_group_a_cg_snapshot_type')
    def test_update_group(self, is_group_a_cg_snapshot_type):
        is_group_a_cg_snapshot_type.return_value = False
        ret = self.driver.update_group(
            self.ctxt, TEST_GROUP[0], add_volumes=[TEST_VOLUME[0]])
        self.assertTupleEqual((None, None, None), ret)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(sqlalchemy_api, 'volume_get', side_effect=_volume_get)
    @mock.patch.object(volume_utils, 'is_group_a_cg_snapshot_type')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_group_snapshot_non_cg(
            self, get_volume_type_qos_specs, is_group_a_cg_snapshot_type,
            volume_get, get_volume_type_extra_specs, request):
        is_group_a_cg_snapshot_type.return_value = False
        extra_specs = {"test1": "aaa"}
        get_volume_type_extra_specs.return_value = extra_specs
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        ret = self.driver.create_group_snapshot(
            self.ctxt, TEST_GROUP_SNAP[0], [TEST_SNAPSHOT[0]]
        )
        self.assertEqual(5, request.call_count)
        actual = (
            {'status': 'available'},
            [{'id': TEST_SNAPSHOT[0]['id'],
              'provider_location': '1',
              'status': 'available'}]
        )
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(requests.Session, "request")
    def test_delete_group_snapshot(self, request):
        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT_PAIR),
                               FakeResponse(200, NOTFOUND_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(200, GET_LDEV_RESULT),
                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
        ret = self.driver.delete_group_snapshot(
            self.ctxt, TEST_GROUP_SNAP[0], [TEST_SNAPSHOT[0]])
        self.assertEqual(14, request.call_count)
        actual = (
            {'status': TEST_GROUP_SNAP[0]['status']},
            [{'id': TEST_SNAPSHOT[0]['id'], 'status': 'deleted'}]
        )
        self.assertTupleEqual(actual, ret)

    @mock.patch.object(requests.Session, "request")
    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
    @mock.patch.object(volume_types, 'get_volume_type_qos_specs')
    def test_create_rep_ldev_and_pair_deduplication_compression(
            self, get_volume_type_qos_specs, get_volume_type_extra_specs,
            request):
        get_volume_type_extra_specs.return_value = {
            'hbsd:topology': 'active_active_mirror_volume',
            'hbsd:capacity_saving': 'deduplication_compression'}
        get_volume_type_qos_specs.return_value = {'qos_specs': None}
        self.snapshot_count = 0

        def _request_side_effect(
                method, url, params, json, headers, auth, timeout, verify):
            if self.configuration.hitachi_storage_id in url:
                if method in ('POST', 'PUT'):
                    return FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if ('/remote-mirror-copygroups' in url or
                            '/journals' in url):
                        return FakeResponse(200, NOTFOUND_RESULT)
                    elif '/remote-mirror-copypairs/' in url:
                        return FakeResponse(
                            200, GET_REMOTE_MIRROR_COPYPAIR_RESULT)
                    elif '/ldevs/' in url:
                        return FakeResponse(200, GET_LDEV_RESULT_REP)
                    elif '/snapshots' in url:
                        if self.snapshot_count < 1:
                            self.snapshot_count = self.snapshot_count + 1
                            return FakeResponse(200, GET_SNAPSHOTS_RESULT)
                        else:
                            return FakeResponse(200, NOTFOUND_RESULT)
            else:
                if method in ('POST', 'PUT'):
                    return FakeResponse(400, REMOTE_COMPLETED_SUCCEEDED_RESULT)
                elif method == 'GET':
                    if '/remote-mirror-copygroups' in url:
                        return FakeResponse(200, NOTFOUND_RESULT)
                    elif '/ldevs/' in url:
                        return FakeResponse(200, GET_LDEV_RESULT_REP)
            if '/ldevs/' in url:
                return FakeResponse(200, GET_LDEV_RESULT_REP)
            else:
                return FakeResponse(
                    200, COMPLETED_SUCCEEDED_RESULT)
        self.driver.common.rep_primary._stats = {}
        self.driver.common.rep_primary._stats['pools'] = [
            {'location_info': {'pool_id': 30}}]
        self.driver.common.rep_secondary._stats = {}
        self.driver.common.rep_secondary._stats['pools'] = [
            {'location_info': {'pool_id': 40}}]
        request.side_effect = _request_side_effect
        self.assertRaises(exception.VolumeDriverException,
                          self.driver.create_cloned_volume,
                          TEST_VOLUME[4],
                          TEST_VOLUME[5])
        self.assertEqual(2, get_volume_type_extra_specs.call_count)
        self.assertEqual(1, get_volume_type_qos_specs.call_count)
        self.assertEqual(14, request.call_count)



#    Copyright (c) 2020 Open-E, Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
from unittest import mock

from oslo_utils import units as o_units

from cinder import context
from cinder import exception
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume.drivers.open_e import iscsi
from cinder.volume.drivers.open_e.jovian_common import exception as jexc
from cinder.volume.drivers.open_e.jovian_common import jdss_common as jcom

UUID_1 = '12345678-1234-1234-1234-000000000001'
UUID_2 = '12345678-1234-1234-1234-000000000002'
UUID_3 = '12345678-1234-1234-1234-000000000003'
UUID_4 = '12345678-1234-1234-1234-000000000004'

UUID_S1 = '12345678-1234-1234-1234-100000000001'
UUID_S2 = '12345678-1234-1234-1234-100000000002'
UUID_S3 = '12345678-1234-1234-1234-100000000003'
UUID_S4 = '12345678-1234-1234-1234-100000000004'

CONFIG_OK = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'jovian_block_size': '128K'
}

CONFIG_BLOCK_SIZE = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'jovian_block_size': '64K'
}

CONFIG_BAD_BLOCK_SIZE = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'jovian_block_size': '61K'
}

CONFIG_BACKEND_NAME = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'volume_backend_name': 'JovianDSS',
    'reserved_percentage': 10,
    'jovian_block_size': '128K'
}

CONFIG_MULTI_HOST = {
    'san_hosts': ['192.168.0.2', '192.168.0.3'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'volume_backend_name': 'JovianDSS',
    'reserved_percentage': 10,
    'jovian_block_size': '128K'
}

SNAPSHOTS_CASCADE_1 = [
    {"name": jcom.sname(UUID_S1, UUID_1),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_1)},
    {"name": jcom.sname(UUID_S1, UUID_2),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_2)},
    {"name": jcom.sname(UUID_S1, UUID_3),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_3)}]

SNAPSHOTS_CASCADE_2 = [
    {"name": jcom.sname(UUID_S1, UUID_1),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_1)},
    {"name": jcom.vname(UUID_2),
     "clones": "Pool-0/" + jcom.vname(UUID_2)},
    {"name": jcom.sname(UUID_S1, UUID_3),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_3)}]

SNAPSHOTS_CASCADE_3 = [
    {"name": jcom.vname(UUID_4),
     "clones": "Pool-0/" + jcom.vname(UUID_4)}]

SNAPSHOTS_EMPTY = []

SNAPSHOTS_CLONE = [
    {"name": jcom.vname(UUID_1),
     "clones": "Pool-0/" + jcom.vname(UUID_1)}]

SNAPSHOTS_GARBAGE = [
    {"name": jcom.sname(UUID_S1, UUID_1),
     "clones": "Pool-0/" + jcom.vname(UUID_2)},
    {"name": jcom.sname(UUID_S1, UUID_2),
     "clones": ""}]

SNAPSHOTS_RECURSIVE_1 = [
    {"name": jcom.sname(UUID_S1, UUID_1),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_1)},
    {"name": jcom.sname(UUID_S1, UUID_2),
     "clones": "Pool-0/" + jcom.hidden(UUID_2)}]

SNAPSHOTS_RECURSIVE_CHAIN_1 = [
    {"name": jcom.sname(UUID_S1, UUID_3),
     "clones": "Pool-0/" + jcom.hidden(UUID_3)}]

SNAPSHOTS_RECURSIVE_CHAIN_2 = [
    {"name": jcom.vname(UUID_2),
     "clones": "Pool-0/" + jcom.hidden(UUID_2)}]


def get_jdss_exceptions():

    out = [jexc.JDSSException(reason="Testing"),
           jexc.JDSSRESTException(request="ra request", reason="Testing"),
           jexc.JDSSRESTProxyException(host="test_host", reason="Testing"),
           jexc.JDSSResourceNotFoundException(res="test_resource"),
           jexc.JDSSVolumeNotFoundException(volume="test_volume"),
           jexc.JDSSSnapshotNotFoundException(snapshot="test_snapshot"),
           jexc.JDSSResourceExistsException(res="test_resource"),
           jexc.JDSSSnapshotExistsException(snapshot="test_snapshot"),
           jexc.JDSSVolumeExistsException(volume="test_volume"),
           jexc.JDSSResourceIsBusyException(res="test_resource"),
           jexc.JDSSSnapshotIsBusyException(snapshot="test_snapshot"),
           jexc.JDSSOSException(message="Some os error")]

    return out


class TestOpenEJovianDSSISCSIDriver(test.TestCase):

    def get_iscsi_driver(self, config):
        ctx = context.get_admin_context()

        cfg = mock.Mock()
        cfg.append_config_values.return_value = None
        cfg.get = lambda val, default: config.get(val, default)

        jdssd = iscsi.JovianISCSIDriver()

        jdssd.configuration = cfg
        lib_to_patch = ('cinder.volume.drivers.open_e.jovian_common.rest.'
                        'JovianRESTAPI')
        with mock.patch(lib_to_patch) as ra:
            ra.is_pool_exists.return_value = True
            jdssd.do_setup(ctx)
        jdssd.ra = mock.Mock()
        jdssd.driver = mock.Mock()
        return jdssd, ctx

    def start_patches(self, patches):
        for p in patches:
            p.start()

    def stop_patches(self, patches):
        for p in patches:
            p.stop()

    def test_check_for_setup_error(self):

        cfg = mock.Mock()
        cfg.append_config_values.return_value = None

        jdssd = iscsi.JovianISCSIDriver()
        jdssd.configuration = cfg

        jdssd.ra = mock.Mock()
        jdssd.driver = mock.Mock()

        # No IP
        jdssd.driver.rest_config_is_ok.return_value = True
        jdssd.jovian_hosts = []
        jdssd.block_size = ['64K']

        self.assertRaises(exception.VolumeDriverException,
                          jdssd.check_for_setup_error)

        # No pool detected
        jdssd.driver.rest_config_is_ok.return_value = False
        jdssd.jovian_hosts = ['192.168.0.2']
        jdssd.block_size = ['64K']

        self.assertRaises(exception.VolumeDriverException,
                          jdssd.check_for_setup_error)
        # Bad block size
        jdssd.driver.rest_config_is_ok.return_value = True
        jdssd.jovian_hosts = ['192.168.0.2', '192.168.0.3']
        jdssd.block_size = ['61K']

        self.assertRaises(exception.InvalidConfigurationValue,
                          jdssd.check_for_setup_error)

    def test_get_provider_info(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)

        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]
        target_name = CONFIG_OK["target_prefix"] + UUID_1
        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )

        jdssd.driver.get_provider_location.return_value = location

        ret = jdssd._get_provider_info(UUID_1)

        jdssd.driver.get_provider_location.assert_called_once_with(UUID_1)
        self.assertEqual(location, ret['provider_location'])
        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])
        self.assertIsNotNone(re.match(cred_format, ret['provider_auth']))

    def test_create_volume(self):

        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vol.size = 1
        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]

        target_name = CONFIG_OK["target_prefix"] + UUID_1
        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )

        jdssd.driver.get_provider_location.return_value = location

        ret = jdssd.create_volume(vol)

        self.assertEqual(location, ret['provider_location'])
        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])
        jdssd.driver.create_volume.assert_called_once_with(
            vol.id, vol.size, sparse=False, block_size="128K")
        self.assertIsNotNone(re.match(cred_format, ret['provider_auth']))
        self.assertEqual(location, ret['provider_location'])

    def test_create_volume_small_block(self):

        jdssd, ctx = self.get_iscsi_driver(CONFIG_BLOCK_SIZE)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        vol.size = 1
        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]

        target_name = CONFIG_OK["target_prefix"] + UUID_1
        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )

        jdssd.driver.create_volume.return_value = None
        jdssd.driver.get_provider_location.return_value = location

        jdssd.ra.get_active_host.return_value = host

        ret = jdssd.create_volume(vol)

        jdssd.driver.create_volume.assert_called_once_with(
            vol.id, vol.size, sparse=False, block_size="64K")

        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )
        self.assertEqual(location, ret['provider_location'])
        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])
        self.assertIsNotNone(re.match(cred_format, ret['provider_auth']))

    def test_delete_volume_cascade(self):
        # Volume with 3 snapshots and 1 clone of a snapshots
        # We should delete childless snapshots
        #           and then cal for volume deletion
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1

        jdssd.driver.delete_volume.return_value = None

        jdssd.delete_volume(vol, cascade=True)

        jdssd.driver.delete_volume.assert_called_once_with(UUID_1,
                                                           cascade=True)

    def test_delete_volume_exceptions(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1

        for exc in get_jdss_exceptions():
            jdssd.driver.delete_volume.side_effect = exc
            try:
                jdssd.delete_volume(vol, cascade=False)
            except Exception as err:
                self.assertIsInstance(err, exception.VolumeBackendAPIException)

    def test_extend_volume(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1
        jdssd.driver.resize_volume.return_value = None
        jdssd.extend_volume(vol, 2)

        jdssd.driver.resize_volume.assert_called_once_with(
            UUID_1, 2)

    def test_extend_volume_exceptions(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1

        for exc in get_jdss_exceptions():
            try:
                jdssd.extend_volume(vol, 2)
            except Exception as err:
                self.assertIsInstance(err, exception.VolumeBackendAPIException)

    def test_create_cloned_volume(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)

        sparse = CONFIG_OK['san_thin_provision']
        orig_vol = fake_volume.fake_volume_obj(ctx)
        orig_vol.id = UUID_1
        orig_vol.size = 1

        clone_vol = fake_volume.fake_volume_obj(ctx)
        clone_vol.id = UUID_2
        clone_vol.size = 1

        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]
        target_name = CONFIG_OK["target_prefix"] + UUID_2

        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )

        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])

        jdssd.driver.get_provider_location.return_value = location
        jdssd.driver.create_cloned_volume.return_value = None

        ret = jdssd.create_cloned_volume(clone_vol, orig_vol)

        jdssd.driver.create_cloned_volume.assert_called_once_with(
            clone_vol.id,
            orig_vol.id,
            clone_vol.size,
            sparse=sparse)

        self.assertEqual(location, ret['provider_location'])
        self.assertIsNotNone(re.match(cred_format, ret['provider_auth']))
        self.assertEqual(location, ret['provider_location'])

    def test_create_volume_from_snapshot(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)

        orig_snap = fake_snapshot.fake_snapshot_obj(ctx)
        orig_snap.id = UUID_S1
        orig_snap.volume_id = UUID_1

        clone_vol = fake_volume.fake_volume_obj(ctx)
        clone_vol.id = UUID_2
        clone_vol.size = 2

        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]
        target_name = CONFIG_OK["target_prefix"] + UUID_2

        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )

        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])

        patches = [
            mock.patch.object(
                jdssd,
                "_get_provider_auth",
                return_value=cred_format)]

        jdssd.driver.get_provider_location.return_value = location
        jdssd.driver.create_cloned_volume.return_value = None

        self.start_patches(patches)

        ret = jdssd.create_volume_from_snapshot(clone_vol, orig_snap)

        jdssd.driver.create_cloned_volume.assert_called_once_with(
            clone_vol.id,
            orig_snap.volume_id,
            clone_vol.size,
            snapshot_name=orig_snap.id)
        self.stop_patches(patches)

        self.assertEqual(location, ret['provider_location'])
        self.assertEqual(cred_format, ret['provider_auth'])

    def test_create_snapshot(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)

        snap = fake_snapshot.fake_snapshot_obj(ctx, id=UUID_S1)
        snap.volume_id = UUID_1

        jdssd.driver.create_snapshot.return_value = None
        jdssd.create_snapshot(snap)

        jdssd.driver.create_snapshot.assert_called_once_with(UUID_S1, UUID_1)

    def test_delete_snapshot(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)

        snap = fake_snapshot.fake_snapshot_obj(ctx,
                                               id=UUID_S1,
                                               volume_id=UUID_1)

        jdssd.driver.delete_snapshot.return_value = None
        jdssd.delete_snapshot(snap)
        jdssd.driver.delete_snapshot.assert_called_once_with(UUID_1, UUID_S1)

    def test_delete_snapshot_exceptions(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)
        snap = fake_snapshot.fake_snapshot_obj(ctx, id=UUID_1)

        for exc in get_jdss_exceptions():
            jdssd.driver.delete_snapshot.side_effect = exc
            try:
                ret = jdssd.delete_snapshot(snap)
                if isinstance(exc, jexc.JDSSVolumeNotFoundException):
                    self.assertTrue(ret is None)
            except Exception as err:
                self.assertIsInstance(err, exception.VolumeBackendAPIException)

    def test_local_path(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)

        vol = fake_snapshot.fake_snapshot_obj(ctx, id=UUID_1)

        self.assertRaises(NotImplementedError, jdssd.local_path, vol)

    def test_get_provider_auth(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)

        auth = jdssd._get_provider_auth()
        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=CONFIG_OK['chap_password_len'])
        self.assertIsNotNone(re.match(cred_format, auth))

    def test_get_provider_auth_long(self):
        long_pass_config = CONFIG_OK.copy()
        long_pass_config['chap_password_len'] = 16
        jdssd, ctx = self.get_iscsi_driver(long_pass_config)

        auth = jdssd._get_provider_auth()
        cred_format = (r"CHAP [0-9,a-z,A-Z]{{{name_len}}} "
                       "[0-9,a-z,A-Z]{{{pass_len}}}").format(
            name_len=8,
            pass_len=16)
        self.assertIsNotNone(re.match(cred_format, auth))

    def test_create_export(self):

        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx, id=UUID_1)

        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]
        target_name = CONFIG_OK["target_prefix"] + UUID_1
        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )

        jdssd.driver.get_provider_location.return_value = location

        jdssd.driver.get_provider_location.return_value = location

        ret = jdssd.create_export(ctx, vol, "connector")
        jdssd.driver.ensure_export.assert_called_once_with(vol.id,
                                                           mock.ANY)

        self.assertEqual(location, ret["provider_location"])

    def test_ensure_export(self):

        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)
        vol.id = UUID_1

        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]
        target_name = CONFIG_OK["target_prefix"] + UUID_1
        location = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )
        auth = 'chap user_name 123456789012'
        jdssd.driver.get_provider_location.return_value = location
        with mock.patch.object(jdssd, '_get_provider_auth'):

            jdssd._get_provider_auth.return_value = auth
            ret = jdssd.ensure_export(ctx, vol)
            jdssd._get_provider_auth.assert_called_once()
        jdssd.driver.ensure_export.assert_called_once_with(vol.id, auth)
        self.assertEqual(location, ret["provider_location"])

    def test_remove_export(self):

        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx, id=UUID_1)

        jdssd.remove_export(ctx, vol)
        jdssd.driver.remove_export.assert_called_once_with(UUID_1)

    def test_update_volume_stats(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_BACKEND_NAME)

        location_info = 'JovianISCSIDriver:192.168.0.2:Pool-0'
        correct_out = {
            'vendor_name': 'Open-E',
            'driver_version': "1.0.3",
            'storage_protocol': 'iSCSI',
            'total_capacity_gb': 100,
            'free_capacity_gb': 50,
            'reserved_percentage': 10,
            'volume_backend_name': CONFIG_BACKEND_NAME['volume_backend_name'],
            'QoS_support': False,
            'location_info': location_info,
            'multiattach': True
        }
        jdssd.ra.get_pool_stats.return_value = {
            'size': 100 * o_units.Gi,
            'available': 50 * o_units.Gi}
        jdssd.ra.get_active_host.return_value = CONFIG_OK['san_hosts'][0]
        jdssd._update_volume_stats()

        self.assertEqual(correct_out, jdssd._stats)

    def test_get_iscsi_properties(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_OK)

        provider_auth = 'chap user_name 123456789012'

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        with mock.patch.object(jdssd.ra, "get_active_host"):
            jdssd.ra.get_active_host.return_value = CONFIG_OK['san_hosts'][0]
            ret = jdssd._get_iscsi_properties(UUID_1, provider_auth,
                                              multipath=False)
        expected = {'auth_method': 'chap',
                    'auth_password': '123456789012',
                    'auth_username': 'user_name',
                    'target_discovered': False,
                    'target_iqn': target_name,
                    'target_lun': 0,
                    'target_portal': '192.168.0.2:3260'}
        self.assertEqual(expected, ret)

    def test_get_iscsi_properties_multipath(self):
        jdssd, ctx = self.get_iscsi_driver(CONFIG_MULTI_HOST)

        provider_auth = 'chap user_name 123456789012'

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        ret = jdssd._get_iscsi_properties(UUID_1,
                                          provider_auth,
                                          multipath=True)
        expected = {'auth_method': 'chap',
                    'auth_password': '123456789012',
                    'auth_username': 'user_name',
                    'target_discovered': False,
                    'target_iqns': [target_name, target_name],
                    'target_lun': 0,
                    'target_luns': [0, 0],
                    'target_portals': ['192.168.0.2:3260', '192.168.0.3:3260']}
        self.assertEqual(expected, ret)

    def test_initialize_connection(self):

        jdssd, ctx = self.get_iscsi_driver(CONFIG_MULTI_HOST)

        vol = fake_volume.fake_volume_obj(ctx, id=UUID_1)
        vol.provider_auth = 'chap user_name 123456789012'

        connector = {'multipath': False, 'ip': '172.16.0.2'}

        target_name = CONFIG_OK['target_prefix'] + UUID_1

        properties = {'auth_method': 'chap',
                      'auth_password': '123456789012',
                      'auth_username': 'user_name',
                      'target_discovered': False,
                      'target_iqn': target_name,
                      'target_lun': 0,
                      'target_portal': '192.168.0.2:3260'}

        con_info = {
            'driver_volume_type': 'iscsi',
            'data': properties,
        }
        jdssd.driver.initialize_connection.return_value = con_info
        init_con_exp = [mock.call(vol.id,
                                  vol.provider_auth,
                                  multipath=False)]
        with mock.patch.object(jdssd.ra, "get_active_host"):
            jdssd.ra.get_active_host.return_value = CONFIG_OK["san_hosts"][0]

            ret = jdssd.initialize_connection(vol, connector)
        jdssd.driver.initialize_connection.assert_has_calls(init_con_exp)
        self.assertEqual(con_info, ret)

    def test_initialize_connection_snapshot(self):

        jdssd, ctx = self.get_iscsi_driver(CONFIG_MULTI_HOST)

        snap = fake_snapshot.fake_snapshot_obj(ctx, id=UUID_1)
        snap.provider_auth = 'chap user_name 123456789012'

        connector = {'multipath': True, 'ip': '172.16.0.2'}

        target_name = CONFIG_OK['target_prefix'] + UUID_1

        properties = {'auth_method': 'chap',
                      'auth_password': '123456789012',
                      'auth_username': 'user_name',
                      'target_discovered': False,
                      'target_iqns': [target_name, target_name],
                      'target_lun': 0,
                      'target_luns': [0, 0],
                      'target_portals': ['192.168.0.2:3260',
                                         '192.168.0.3:3260']}

        con_info = {
            'driver_volume_type': 'iscsi',
            'data': properties,
        }
        init_con_exp = [mock.call(snap.volume_id,
                                  snap.provider_auth,
                                  snapshot_id=snap.id,
                                  multipath=True)]

        jdssd.driver.initialize_connection.return_value = con_info
        ret = jdssd.initialize_connection_snapshot(snap, connector)

        jdssd.driver.initialize_connection.assert_has_calls(init_con_exp)
        self.assertEqual(con_info, ret)
